﻿using System;
using System.Diagnostics;
using System.Net;
using System.IO;
using System.Windows.Forms;
using PCSC;
using PCSC.Iso7816;
using System.Text.Json.Serialization;
using Newtonsoft.Json.Linq;
using Newtonsoft.Json;
using System.Collections.Generic;
using System.Text;

namespace VivokeyACR
{
    public partial class Form1 : Form
    {
        private const string baseURL = "https://api2.vivokey.com/v1/";
        private const string apikey = "";
        private const string getChallengeEndpoint = "get-challenge";
        private const string pcdChallengeEndpoint = "pcd-challenge";
        private const string checkResponseEndpoint = "check-response";

        ContextFactory contextFactory = new ContextFactory();

        public struct AUTH
        {
            public string PICCUID;
            public string PICCchallenge;
            public string PICCresponse;
            public string PCDchallenge;
            public string PCDresponse;

            public string pcdResponse()
            {
                Dictionary<string, string> post = new Dictionary<string, string>
                {
                    {"picc-uid", PICCUID },
                    {"picc-challenge", PICCchallenge },
                    {"pcd-challenge", PCDchallenge }
                };

                return JsonConvert.SerializeObject(post);
            }
        }


        public Form1()
        {
            InitializeComponent();            
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            refreshReaders();

            label1.Text = "PICC-UID: ";
            label2.Text = "PICC-challenge: ";
            label3.Text = "PCD-challenge: ";
            label4.Text = "PCD-response: ";
            label5.Text = "PICC-response: ";
        }

        private void button1_Click(object sender, EventArgs e)
        {
            refreshReaders();
        }

        private void button2_Click(object sender, EventArgs e)
        {

            label1.Text = "PICC-UID: ";
            label2.Text = "PICC-challenge: ";
            label3.Text = "PCD-challenge: ";
            label4.Text = "PCD-response: ";
            label5.Text = "PICC-response: ";

            AUTH auth = new AUTH();

            auth.PICCUID = getUID();

            label1.Text = label1.Text + auth.PICCUID;

            auth.PCDchallenge = getPCDChallenge();

            label3.Text = label3.Text + auth.PCDchallenge;

            auth.PICCchallenge = getPICCChallenge().ToUpper();

            label2.Text = label2.Text + auth.PICCchallenge;

            auth.PCDresponse = getPCDResponse(auth);

            label4.Text = label4.Text + auth.PCDresponse;

            if (auth.PCDresponse != "error")
            {
                auth.PICCresponse = sendPCDResponse(auth);

                label5.Text = label5.Text + auth.PICCresponse;
            }
        }

        private void button3_Click(object sender, EventArgs e)
        {
            using (var ctx = contextFactory.Establish(SCardScope.System))
            {
                using (var isoReader = new IsoReader(ctx, comboBox1.Text, SCardShareMode.Shared, SCardProtocol.Any, false))
                {

                    var apdu = new CommandApdu(IsoCase.Case4Short, isoReader.ActiveProtocol)
                    {
                        CLA = 0xFF, // Class
                        INS = 0x00,
                        P1 = 0x40,
                        P2 = 0xAA,
                        Data = new byte[] { 0x01, 0x01, 0x03, 0x00},
                        Le = 0x00
                    };

                    var response = isoReader.Transmit(apdu);
                }
            }
        }

        private string getPICCChallenge()
        {
            string challenge;
            var httpWebRequest = (HttpWebRequest)WebRequest.Create(baseURL + getChallengeEndpoint);
            httpWebRequest.ContentType = "application/json";
            httpWebRequest.Method = "POST";

            using (var streamWriter = new StreamWriter(httpWebRequest.GetRequestStream()))
            {
                Dictionary<string, string> post = new Dictionary<string, string>
                {
                    {"api-key", apikey }
                };

                string json = JsonConvert.SerializeObject(post);

                streamWriter.Write(json);
            }

            var httpResponse = (HttpWebResponse)httpWebRequest.GetResponse();
            using (var streamReader = new StreamReader(httpResponse.GetResponseStream()))
            {
                challenge = streamReader.ReadToEnd();
            }

            var response = JObject.Parse(challenge);
            var piccchallenge = response["picc-challenge"];

            return piccchallenge.ToString();

        }

        private string getUID()
        {
            using (var ctx = contextFactory.Establish(SCardScope.System))
            {
                using (var isoReader = new IsoReader(ctx, comboBox1.Text, SCardShareMode.Shared, SCardProtocol.Any, false))
                {

                    var apdu = new CommandApdu(IsoCase.Case2Short, isoReader.ActiveProtocol)
                    {
                        CLA = 0xFF, // Class
                        Instruction = InstructionCode.GetData,
                        P1 = 0x00, // Parameter 1
                        P2 = 0x00, // Parameter 2
                        Le = 0x07 // Expected length of the returned data
                    };

                    var response = isoReader.Transmit(apdu);
                    var array = response.GetData();

                    return BitConverter.ToString(array).Replace("-", string.Empty);
                }
            }
        }

        private string getPCDChallenge()
        {
            using (var ctx = contextFactory.Establish(SCardScope.System))
            {
                using (var isoReader = new IsoReader(ctx, comboBox1.Text, SCardShareMode.Shared, SCardProtocol.Any, false))
                {

                    var apdu = new CommandApdu(IsoCase.Case4Short, isoReader.ActiveProtocol)
                    {
                        CLA = 0x90,
                        INS = 0x71,
                        P1 = 0x00, 
                        P2 = 0x00, 
                        Data = new byte[] { 0x02, 0x00 },
                        Le = 0x00
                    };

                    var response = isoReader.Transmit(apdu);
                    var array = response.GetData();

                    return BitConverter.ToString(array).Replace("-", string.Empty);
                }
            }
        }

        private string getPCDResponse(AUTH auth)
        {
            string jsonResponse;
            var httpWebRequest = (HttpWebRequest)WebRequest.Create(baseURL + pcdChallengeEndpoint);
            httpWebRequest.ContentType = "application/json";
            httpWebRequest.Method = "POST";

            using (var streamWriter = new StreamWriter(httpWebRequest.GetRequestStream()))
            { 
                streamWriter.Write(auth.pcdResponse());
            }

            var httpResponse = (HttpWebResponse)httpWebRequest.GetResponse();
            using (var streamReader = new StreamReader(httpResponse.GetResponseStream()))
            {
                jsonResponse = streamReader.ReadToEnd();
            }

            var response = JObject.Parse(jsonResponse);
            var pcdchallenge = response["pcd-response"];

            return pcdchallenge.ToString();
        }

        private string sendPCDResponse(AUTH auth)
        {
            byte[] data = Encoding.ASCII.GetBytes(auth.PCDresponse);

            using (var ctx = contextFactory.Establish(SCardScope.System))
            {
                using (var isoReader = new IsoReader(ctx, comboBox1.Text, SCardShareMode.Shared, SCardProtocol.Any, false))
                {

                    var apdu = new CommandApdu(IsoCase.Case4Short, isoReader.ActiveProtocol)
                    {
                        CLA = 0x90, // Class
                        INS = 0xaf,
                        P1 = 0x00, // Parameter 1
                        P2 = 0x00, // Parameter 2
                        Data = data,
                        Le = 0x00
                    };

                    var response = isoReader.Transmit(apdu);

                    return BitConverter.ToString(response.GetData()).Replace("-", string.Empty);
                }
            }
        }

        private void refreshReaders()
        {
            using (var context = contextFactory.Establish(SCardScope.System))
            {
                comboBox1.Items.Clear();
                var readerNames = context.GetReaders();
                foreach (var readerName in readerNames)
                {
                    comboBox1.Items.Add(readerName);
                    comboBox1.SelectedIndex = 0;
                }
            }
        }

        public static string Reverse(string s)
        {
            char[] charArray = s.ToCharArray();
            Array.Reverse(charArray);
            return new string(charArray);
        }

    }
}
